static char *cvs_id =" $Id: mygets.c,v 1.5 2003/02/25 09:49:32 berar Exp $ ";
/* mygets.c               minimal readline function set
 *
 * jf berar d2am, berar@esrf.fr
 *           21Feb03   :  minimal readline function set
 */
/* ------------------------------------------------------ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <termios.h>
#include <unistd.h>
/* ------------------------------------------------------ */
#define NDEBUG 1
/* ------------------------------------------------------ */
#define HIST_DEPTH 5
#define BUFF_SIZE 256
/* ------------------------------------------------------ */
static char **history=NULL;
static int  hist_depth=0;
static int  hist_set=0;
static int  hist_get=0;
static char buffer[BUFF_SIZE-1];
static int  term_in=-1;
static int  term_out=-1;
static int  term_insert=1;
static int  backspace_len=1;
static int  escape_len=1;
struct termios termios_in;
struct termios termios_out;
static char *fileline (void);
static char *termline (void);
static void clean_term (const int len);
static int  move_back (int pos, char *buf, int *len);
static int  move_next (int pos, char *buf, int *len);
static int  rewrite_char (int pos, char *buf, int *len);
static int  insert_char (const char t, int pos, char *buf, int *len);
static int  delete_char (int pos, char *buf, int *len);
static int  delete_pchar (int pos, char *buf, int *len);
static void history_save(char *buf, int len);
static int  history_read(char *buf, int *len, int dir);
/* ------------------------------------------------------ */
void  readline_exit(void)
{   if (term_in!=-1)  tcsetattr(term_in, TCSANOW, &termios_in);
    if (term_out!=-1) tcsetattr(term_out, TCSANOW, &termios_out);
    }
/* ------------------------------------------------------ */
void  readline_init(const int depth)
{   struct termios term_mod;
    hist_depth=depth>0?depth:HIST_DEPTH;
    history= (char**) malloc(hist_depth*sizeof(char*));
    memset(history,0,hist_depth*sizeof(char*));
    if (tcgetattr(fileno(stdin), &termios_in)) return;
    term_in=fileno(stdin);
    if (tcgetattr(fileno(stdout), &termios_out)==0) 
       term_out=fileno(stdout);
    atexit(readline_exit); 
    if (termios_out.c_lflag & ICANON)
    {   termios_out.c_lflag=termios_out.c_lflag & ~ICANON;
        tcsetattr(term_out, TCSANOW, &termios_out);
        termios_out.c_lflag=termios_out.c_lflag | ICANON;
        }
#ifdef XDEBUG
printf("DEBUG STDIN=%d iflag %4X  lflag %4X\n",term_in,termios_in.c_iflag,termios_in.c_lflag);
printf("DEBUG ICANON(%X) %X ECHOCTL(%X) %4X ECHOE(%X) %4X\n",
   ICANON, termios_in.c_lflag & ICANON, 
   ECHOCTL, termios_in.c_lflag & ECHOCTL,
   ECHOE, termios_in.c_lflag & ECHOE);
{ int i; char c,d,e ;
 for (i=0; i<NCCS; i++) printf("CC[%d]=%X  %s",i, termios_in.c_cc[i],i%4==3?"\n":" ");
 printf("\n TEST a^Hb=");c=(char) getchar(); d=(char) getchar(); e=(char) getchar();
 printf("\n--%c==%c--==%c\n",c,d,e);
 printf("\n TEST a'del'b=");c=(char) getchar(); d=(char) getchar(); e=(char) getchar();
 printf("\n--%c==%c--==%c\n",c,d,e);
 }
#endif
    if (termios_in.c_lflag & ECHOCTL )
    {   if (ECHOCTL==0x200) backspace_len=2; /*linux*/
        else if (ECHOCTL==0x100) backspace_len=-1; /*hpux*/
        else backspace_len=2;
        escape_len=4;
        }
    }
/* ------------------------------------------------------ */
char *readline (const char * prompt)
{  char *c, *d;
   if (!hist_depth) readline_init (HIST_DEPTH);
   if (prompt) printf(prompt);
   if ( term_in>=0 ) return termline ();
   else return fileline ();
   }  
/* ------------------------------------------------------ */
static char *fileline (void)
{  char *c, *d;
   if (feof(stdin)) { printf("EOF\n"); return "q";}
   *buffer=0;
   fgets(buffer,BUFF_SIZE,stdin);
   if (*buffer==0 & feof(stdin)) { printf("EOF\n"); return "q";}
   if (ferror(stdin)) { printf("ERROR\n"); return "q";}
   if ((c=strchr(buffer,'\n'))) *c=0;
   c=buffer;
   printf("%s\n",c);
	while ( *c == ' ' ) c++;
   d=(char *)malloc(strlen(c)+1);
   strcpy(d,c);
   return d;
   }  
/* ------------------------------------------------------ */
static char *termline (void)
{  int l,f;
   char t, *c, *d;
   f=l=0;
   do
   {  t=(char) getchar();
      if ( t=='\n') break;
      switch (t)
      {  case 0x08 : clean_term(backspace_len);
                     l=delete_pchar(l, buffer, &f);
                     break;
         case 0x7F : clean_term(backspace_len);
                     l=delete_char(l, buffer, &f);
                     break;
         case 0x1B : {  char u;
                        t=(char) getchar();
                        if (t!='[') break;
                        u=(char) getchar();
                        if (u=='A') /* up */
                        {  clean_term(escape_len);
                           clean_term(l);
                           l=history_read(buffer,&f,1); 
                           break;
                           }
                        if (u=='B') /* down */
                        {  clean_term(escape_len);
                           clean_term(l);
                           l=history_read(buffer,&f,-1); 
                           break;
                           }
                        if (u=='C') /* forward */
                        {  clean_term(escape_len);
                           l=move_next(l, buffer, &f);
                           break;
                           }
                        if (u=='D')  /* back */
                        {  clean_term(escape_len);
                           l=move_back(l, buffer, &f);
                           break;
                           }
                        if (u=='2' && (char)getchar()=='~')/* insert */
                        {  clean_term(escape_len+1);
                           l=rewrite_char(l, buffer, &f);
                           term_insert=!term_insert;
                           }
                        if (u=='3' && (char)getchar()=='~')/* del */
                        {  clean_term(escape_len+1);
                           l=delete_char(l, buffer, &f);
                           }
                        break;
                        }
         default :  if (f==l) { buffer[l++]=t; f++; }
                    else l=insert_char(t, l, buffer, &f);
         }
      } while ((f<BUFF_SIZE) && (t !='\n'));
   buffer[f]=0;
   history_save(buffer,f);
#ifdef DEBUG
   printf("BUFFER %02X %02X %02X %02X =%s\n", buffer[0],buffer[1],buffer[2],buffer[3], buffer);
#endif
   c=buffer;
	while ( *c == ' ' || *c == 0xFF ) c++;  /* 0xFF pour HP */
   d=(char *)malloc(strlen(c)+1);
   strcpy(d,c);
   return d;
   }  
/* ------------------------------------------------------ */
static void clean_term(const int len)
{  int i;
   if (len>0)
   { for (i=0;i<len;i++) putchar('\b');
     for (i=0;i<len;i++) putchar(' ');
     for (i=0;i<len;i++) putchar('\b');
     }
   else if (len<0)
     for (i=0;i<-len;i++) putchar(' ');
   }  
/* ------------------------------------------------------ */
static int move_back(int pos, char *buf, int *len)
{  int i;
   if (*len==0) return 0;
   if (pos>*len) pos=*len;
   if (pos>0)  {  pos--; putchar('\b'); }
   for (i=pos;i<*len;i++) putchar(buf[i]);
   for (i=*len;i>pos;i--) putchar('\b');
   return pos;
   }  
/* ------------------------------------------------------ */
static int move_next(int pos, char *buf, int *len)
{  int i;
   if (pos>=*len) return pos;
   for (i=pos;i<=*len;i++) putchar(buf[i]);
   for (i=*len;i>pos+1;i--) putchar('\b');
   return pos+1;
   }  
/* ------------------------------------------------------ */
static int rewrite_char(int pos, char *buf, int *len)
{  int i;
   if (pos>=*len) return pos;
   for (i=pos;i<=*len;i++) putchar(buf[i]);
   for (i=*len;i>pos;i--) putchar('\b');
   return pos;
   }  
/* ------------------------------------------------------ */
static int insert_char(const char t, int pos, char *buf, int *len)
{  int i;
   if (pos>*len) pos=*len;
   if (term_insert)
   {  *len+=1;
      for(i=*len;i>pos;i--) buf[i]=buf[i-1];
      }
   buf[pos]=t;
   putchar('\b');
   for (i=pos;i<*len;i++) putchar(buf[i]);
   for (i=*len;i>pos;i--) putchar('\b');
   return pos;
   }  
/* ------------------------------------------------------ */
static int  delete_char (int pos, char *buf, int *len)
{  int i;
   if (*len==0) return 0;
   if (pos>=*len-1)
   {  putchar('\b'); putchar(' '); putchar('\b');
      *len-=1;
      return *len;
      }
   *len-=1;
   for(i=pos;i<*len;i++) buf[i]=buf[i+1];
   for (i=pos;i<*len;i++) putchar(buf[i]);
   putchar(' ');
   for (i=*len+1;i>pos;i--) putchar('\b');
   return pos;
   }  
/* ------------------------------------------------------ */
static int  delete_pchar (int pos, char *buf, int *len)
{  int i;
   if (*len==0) return 0;
   if (pos>=*len-1)
   {  putchar('\b'); putchar(' '); putchar('\b');
      *len-=1;
      return *len;
      }
   *len-=1;
   putchar('\b');
   for(i=pos-1;i<*len;i++) buf[i]=buf[i+1];
   for (i=pos-1;i<*len;i++) putchar(buf[i]);
   putchar(' ');
   for (i=*len+1;i>=pos;i--) putchar('\b');
   return pos-1;
   }  
/* ------------------------------------------------------ */
static void history_save( char *buf, int len)
{  if (history[hist_get] && strcmp(history[hist_get],buf)==0) return;
   if (history[hist_set])  free(history[hist_set]);
   history[hist_set]=(char *)malloc(len+1);
   strcpy(history[hist_set],buf);
   if (++hist_set >= hist_depth) hist_set=0;
   hist_get=hist_set;
   }  
/* ------------------------------------------------------ */
static int history_read( char *buf, int *len, int dir)
{  int i;
   hist_get=(hist_get-dir)%hist_depth;
   if (hist_get<0) hist_get=hist_depth-1;
   if (history[hist_get]) 
   {  strcpy(buf, history[hist_get]);
      printf(buf);
      }
   else  strcpy(buf, "");
   *len=strlen(buf);
   return *len;
   }  
/* ------------------------------------------------------ */
